@@ -6,7 +6,7 @@
  *	License as published by the Free Software Foundation version 2.1
  *	of the License.
  *
- * Copyright (c) 2003-2010 Thomas Graf <tgraf@suug.ch>
+ * Copyright (c) 2003-2013 Thomas Graf <tgraf@suug.ch>
  */
 
 /**
@@ -38,10 +38,12 @@
 #define VLAN_HAS_FLAGS		(1<<1)
 #define VLAN_HAS_INGRESS_QOS	(1<<2)
 #define VLAN_HAS_EGRESS_QOS	(1<<3)
+#define VLAN_HAS_PROTOCOL	(1<<4)
 
 struct vlan_info
 {
 	uint16_t		vi_vlan_id;
+	uint16_t		vi_protocol;
 	uint32_t		vi_flags;
 	uint32_t		vi_flags_mask;
 	uint32_t		vi_ingress_qos[VLAN_PRIO_MAX+1];
@@ -58,16 +60,23 @@ static struct nla_policy vlan_policy[IFLA_VLAN_MAX+1] = {
 	[IFLA_VLAN_FLAGS]	= { .minlen = sizeof(struct ifla_vlan_flags) },
 	[IFLA_VLAN_INGRESS_QOS]	= { .type = NLA_NESTED },
 	[IFLA_VLAN_EGRESS_QOS]	= { .type = NLA_NESTED },
+	[IFLA_VLAN_PROTOCOL]	= { .type = NLA_U16 },
 };
 
 static int vlan_alloc(struct rtnl_link *link)
 {
 	struct vlan_info *vi;
 
-	if ((vi = calloc(1, sizeof(*vi))) == NULL)
-		return -NLE_NOMEM;
+	if (link->l_info) {
+		vi = link->l_info;
+		free(vi->vi_egress_qos);
+		memset(link->l_info, 0, sizeof(*vi));
+	} else {
+		if ((vi = calloc(1, sizeof(*vi))) == NULL)
+			return -NLE_NOMEM;
 
-	link->l_info = vi;
+		link->l_info = vi;
+	}
 
 	return 0;
 }
@@ -94,6 +103,11 @@ static int vlan_parse(struct rtnl_link *link, struct nlattr *data,
 		vi->vi_mask |= VLAN_HAS_ID;
 	}
 
+	if (tb[IFLA_VLAN_PROTOCOL]) {
+		vi->vi_protocol = nla_get_u16(tb[IFLA_VLAN_PROTOCOL]);
+		vi->vi_mask |= VLAN_HAS_PROTOCOL;
+	}
+
 	if (tb[IFLA_VLAN_FLAGS]) {
 		struct ifla_vlan_flags flags;
 		nla_memcpy(&flags, tb[IFLA_VLAN_FLAGS], sizeof(flags));
@@ -137,7 +151,7 @@ static int vlan_parse(struct rtnl_link *link, struct nlattr *data,
 
 		/* align to have a little reserve */
 		vi->vi_egress_size = (i + 32) & ~31;
-		vi->vi_egress_qos = calloc(vi->vi_egress_size, sizeof(*map));
+		vi->vi_egress_qos = calloc(vi->vi_egress_size, sizeof(*vi->vi_egress_qos));
 		if (vi->vi_egress_qos == NULL)
 			return -NLE_NOMEM;
 
@@ -186,7 +200,12 @@ static void vlan_dump_details(struct rtnl_link *link, struct nl_dump_params *p)
 	char buf[64];
 
 	rtnl_link_vlan_flags2str(vi->vi_flags, buf, sizeof(buf));
-	nl_dump_line(p, "    vlan-info id %d <%s>\n", vi->vi_vlan_id, buf);
+	nl_dump_line(p, "    vlan-info id %d <%s>", vi->vi_vlan_id, buf);
+
+	if (vi->vi_mask & VLAN_HAS_PROTOCOL)
+		nl_dump_line(p, "    vlan protocol <%d>", vi->vi_protocol);
+
+	nl_dump(p, "\n");
 
 	if (vi->vi_mask & VLAN_HAS_INGRESS_QOS) {
 		nl_dump_line(p, 
@@ -340,6 +359,27 @@ static struct rtnl_link_info_ops vlan_info_ops = {
  * @{
  */
 
+/**
+ * Allocate link object of type VLAN
+ *
+ * @return Allocated link object or NULL.
+ */
+struct rtnl_link *rtnl_link_vlan_alloc(void)
+{
+	struct rtnl_link *link;
+	int err;
+
+	if (!(link = rtnl_link_alloc()))
+		return NULL;
+
+	if ((err = rtnl_link_set_type(link, "vlan")) < 0) {
+		rtnl_link_put(link);
+		return NULL;
+	}
+
+	return link;
+}
+
 /**
  * Check if link is a VLAN link
  * @arg link		Link object
@@ -388,6 +428,43 @@ int rtnl_link_vlan_get_id(struct rtnl_link *link)
 		return 0;
 }
 
+/**
+ * Set VLAN protocol
+ * @arg link		Link object
+ * @arg protocol	VLAN protocol
+ *
+ * @return 0 on success or a negative error code
+ */
+int rtnl_link_vlan_set_protocol(struct rtnl_link *link, uint16_t protocol)
+{
+	struct vlan_info *vi = link->l_info;
+
+	IS_VLAN_LINK_ASSERT(link);
+
+	vi->vi_protocol = protocol;
+	vi->vi_mask |= VLAN_HAS_PROTOCOL;
+
+	return 0;
+}
+
+/**
+ * Get VLAN protocol
+ * @arg link		Link object
+ *
+ * @return VLAN protocol, 0 if not set or a negative error code.
+ */
+int rtnl_link_vlan_get_protocol(struct rtnl_link *link)
+{
+	struct vlan_info *vi = link->l_info;
+
+	IS_VLAN_LINK_ASSERT(link);
+
+	if (vi->vi_mask & VLAN_HAS_PROTOCOL)
+		return vi->vi_protocol;
+	else
+		return 0;
+}
+
 /**
  * Set VLAN flags
  * @arg link		Link object
@@ -532,7 +609,7 @@ struct vlan_map *rtnl_link_vlan_get_egress_map(struct rtnl_link *link,
 /** @} */
 
 static const struct trans_tbl vlan_flags[] = {
-	__ADD(VLAN_FLAG_REORDER_HDR, reorder_hdr)
+	__ADD(VLAN_FLAG_REORDER_HDR, reorder_hdr),
 };
 
 /**
