@@ -34,15 +34,20 @@
  * Fills out a routing netlink request message and sends it out
  * using nl_send_simple().
  *
- * @return 0 on success or a negative error code.
+ * @return 0 on success or a negative error code. Due to a bug in
+ * older versions, this returned the number of bytes sent. So for
+ * compatibility, treat positive return values as success too.
  */
 int nl_rtgen_request(struct nl_sock *sk, int type, int family, int flags)
 {
+	int err;
 	struct rtgenmsg gmsg = {
 		.rtgen_family = family,
 	};
 
-	return nl_send_simple(sk, type, flags, &gmsg, sizeof(gmsg));
+	err = nl_send_simple(sk, type, flags, &gmsg, sizeof(gmsg));
+
+	return err >= 0 ? 0 : err;
 }
 
 /** @} */
@@ -53,18 +58,18 @@ int nl_rtgen_request(struct nl_sock *sk, int type, int family, int flags)
  */
 
 static const struct trans_tbl rtntypes[] = {
-	__ADD(RTN_UNSPEC,unspec)
-	__ADD(RTN_UNICAST,unicast)
-	__ADD(RTN_LOCAL,local)
-	__ADD(RTN_BROADCAST,broadcast)
-	__ADD(RTN_ANYCAST,anycast)
-	__ADD(RTN_MULTICAST,multicast)
-	__ADD(RTN_BLACKHOLE,blackhole)
-	__ADD(RTN_UNREACHABLE,unreachable)
-	__ADD(RTN_PROHIBIT,prohibit)
-	__ADD(RTN_THROW,throw)
-	__ADD(RTN_NAT,nat)
-	__ADD(RTN_XRESOLVE,xresolve)
+	__ADD(RTN_UNSPEC,unspec),
+	__ADD(RTN_UNICAST,unicast),
+	__ADD(RTN_LOCAL,local),
+	__ADD(RTN_BROADCAST,broadcast),
+	__ADD(RTN_ANYCAST,anycast),
+	__ADD(RTN_MULTICAST,multicast),
+	__ADD(RTN_BLACKHOLE,blackhole),
+	__ADD(RTN_UNREACHABLE,unreachable),
+	__ADD(RTN_PROHIBIT,prohibit),
+	__ADD(RTN_THROW,throw),
+	__ADD(RTN_NAT,nat),
+	__ADD(RTN_XRESOLVE,xresolve),
 };
 
 char *nl_rtntype2str(int type, char *buf, size_t size)
@@ -85,11 +90,11 @@ int nl_str2rtntype(const char *name)
  */
 
 static const struct trans_tbl scopes[] = {
-	__ADD(255,nowhere)
-	__ADD(254,host)
-	__ADD(253,link)
-	__ADD(200,site)
-	__ADD(0,global)
+	__ADD(255,nowhere),
+	__ADD(254,host),
+	__ADD(253,link),
+	__ADD(200,site),
+	__ADD(0,global),
 };
 
 char *rtnl_scope2str(int scope, char *buf, size_t size)
