@@ -29,7 +29,10 @@
 
 #include "defs.h"
 
+#include "sys/socket.h"
+
 #include <netlink-private/netlink.h>
+#include <netlink-private/socket.h>
 #include <netlink/netlink.h>
 #include <netlink/utils.h>
 #include <netlink/handlers.h>
@@ -61,16 +64,39 @@ static NL_RW_LOCK(port_map_lock);
 
 static uint32_t generate_local_port(void)
 {
-	int i, n;
+	int i, j, n, m;
+	static uint16_t idx_state = 0;
 	uint32_t pid = getpid() & 0x3FFFFF;
 
 	nl_write_lock(&port_map_lock);
 
-	for (i = 0; i < 32; i++) {
+	if (idx_state == 0) {
+		uint32_t t = time(NULL);
+
+		/* from time to time (on average each 2^15 calls), the idx_state will
+		 * be zero again. No problem, just "seed" anew with time(). */
+		idx_state = t ^ (t >> 16) ^ 0x3047;
+	} else
+		idx_state = idx_state + 20011; /* add prime number */
+
+	i = idx_state >> 5;
+	n = idx_state;
+	for (j = 0; j < 32; j++) {
+		/* walk the index somewhat randomized, with always leaving the block
+		 * #0 as last. The reason is that libnl-1 will start at block #0,
+		 * so just leave the first 32 ports preferably for libnl-1 owned sockets
+		 * (this is relevant only if the applications ends up using both versions
+		 * of the library and doesn't hurt otherwise). */
+		if (j == 31)
+			i = 0;
+		else
+			i = (((i-1) + 7) % 31) + 1;
+
 		if (used_ports_map[i] == 0xFFFFFFFF)
 			continue;
 
-		for (n = 0; n < 32; n++) {
+		for (m = 0; m < 32; m++) {
+			n = (n + 13) % 32;
 			if (1UL & (used_ports_map[i] >> n))
 				continue;
 
@@ -82,30 +108,73 @@ static uint32_t generate_local_port(void)
 
 			nl_write_unlock(&port_map_lock);
 
-			return pid + (n << 22);
+			return pid + (((uint32_t)n) << 22);
 		}
 	}
 
 	nl_write_unlock(&port_map_lock);
 
 	/* Out of sockets in our own PID namespace, what to do? FIXME */
-	return UINT_MAX;
+	NL_DBG(1, "Warning: Ran out of unique local port namespace\n");
+	return UINT32_MAX;
 }
 
 static void release_local_port(uint32_t port)
 {
 	int nr;
+	uint32_t mask;
 
-	if (port == UINT_MAX)
+	if (port == UINT32_MAX)
 		return;
-	
+
+	BUG_ON(port == 0);
+
 	nr = port >> 22;
+	mask = 1UL << (nr % 32);
+	nr /= 32;
 
 	nl_write_lock(&port_map_lock);
-	used_ports_map[nr / 32] &= ~(1 << (nr % 32));
+	BUG_ON((used_ports_map[nr] & mask) != mask);
+	used_ports_map[nr] &= ~mask;
 	nl_write_unlock(&port_map_lock);
 }
 
+/** \cond skip */
+void _nl_socket_used_ports_release_all(const uint32_t *used_ports)
+{
+	int i;
+
+	for (i = 0; i < 32; i++) {
+		if (used_ports[i] != 0) {
+			nl_write_lock(&port_map_lock);
+			for (; i < 32; i++) {
+				BUG_ON((used_ports_map[i] & used_ports[i]) != used_ports[i]);
+				used_ports_map[i] &= ~(used_ports[i]);
+			}
+			nl_write_unlock(&port_map_lock);
+			return;
+		}
+	}
+}
+
+void _nl_socket_used_ports_set(uint32_t *used_ports, uint32_t port)
+{
+	int nr;
+	int32_t mask;
+
+	nr = port >> 22;
+	mask = 1UL << (nr % 32);
+	nr /= 32;
+
+	/*
+	BUG_ON(port == UINT32_MAX || port == 0 || (getpid() & 0x3FFFFF) != (port & 0x3FFFFF));
+	BUG_ON(used_ports[nr] & mask);
+	*/
+
+	used_ports[nr] |= mask;
+}
+/** \endcond */
+
 /**
  * @name Allocation
  * @{
@@ -120,15 +189,13 @@ static struct nl_sock *__alloc_socket(struct nl_cb *cb)
 		return NULL;
 
 	sk->s_fd = -1;
-	sk->s_cb = cb;
+	sk->s_cb = nl_cb_get(cb);
 	sk->s_local.nl_family = AF_NETLINK;
 	sk->s_peer.nl_family = AF_NETLINK;
-	sk->s_seq_expect = sk->s_seq_next = time(0);
-	sk->s_local.nl_pid = generate_local_port();
-	if (sk->s_local.nl_pid == UINT_MAX) {
-		nl_socket_free(sk);
-		return NULL;
-	}
+	sk->s_seq_expect = sk->s_seq_next = time(NULL);
+
+	/* the port is 0 (unspecified), meaning NL_OWN_PORT */
+	sk->s_flags = NL_OWN_PORT;
 
 	return sk;
 }
@@ -141,12 +208,18 @@ static struct nl_sock *__alloc_socket(struct nl_cb *cb)
 struct nl_sock *nl_socket_alloc(void)
 {
 	struct nl_cb *cb;
-	
+        struct nl_sock *sk;
+
 	cb = nl_cb_alloc(default_cb);
 	if (!cb)
 		return NULL;
 
-	return __alloc_socket(cb);
+        /* will increment cb reference count on success */
+	sk = __alloc_socket(cb);
+
+        nl_cb_put(cb);
+
+        return sk;
 }
 
 /**
@@ -163,7 +236,7 @@ struct nl_sock *nl_socket_alloc_cb(struct nl_cb *cb)
 	if (cb == NULL)
 		BUG();
 
-	return __alloc_socket(nl_cb_get(cb));
+	return __alloc_socket(cb);
 }
 
 /**
@@ -258,6 +331,26 @@ void nl_socket_enable_auto_ack(struct nl_sock *sk)
 
 /** @} */
 
+/** \cond skip */
+int _nl_socket_is_local_port_unspecified(struct nl_sock *sk)
+{
+	return (sk->s_local.nl_pid == 0);
+}
+
+uint32_t _nl_socket_generate_local_port_no_release(struct nl_sock *sk)
+{
+	uint32_t port;
+
+	/* reset the port to generate_local_port(), but do not release
+	 * the previously generated port. */
+
+	port = generate_local_port();
+	sk->s_flags &= ~NL_OWN_PORT;
+	sk->s_local.nl_pid = port;
+	return port;
+}
+/** \endcond */
+
 /**
  * @name Source Idenficiation
  * @{
@@ -265,6 +358,18 @@ void nl_socket_enable_auto_ack(struct nl_sock *sk)
 
 uint32_t nl_socket_get_local_port(const struct nl_sock *sk)
 {
+	if (sk->s_local.nl_pid == 0) {
+		/* modify the const argument sk. This is justified, because
+		 * nobody ever saw the local_port from externally. So, we
+		 * initilize it on first use.
+		 *
+		 * Note that this also means that you cannot call this function
+		 * from multiple threads without synchronization. But nl_sock
+		 * is not automatically threadsafe anyway, so the user is not
+		 * allowed to do that.
+		 */
+		return _nl_socket_generate_local_port_no_release((struct nl_sock *) sk);
+	}
 	return sk->s_local.nl_pid;
 }
 
@@ -273,27 +378,18 @@ uint32_t nl_socket_get_local_port(const struct nl_sock *sk)
  * @arg sk		Netlink socket.
  * @arg port		Local port identifier
  *
- * Assigns a local port identifier to the socket. If port is 0
- * a unique port identifier will be generated automatically.
+ * Assigns a local port identifier to the socket.
+ *
+ * If port is 0, the port is reset to 'unspecified' as it is after newly
+ * calling nl_socket_alloc().
+ * Unspecified means, that the port will be generated automatically later
+ * on first use (either on nl_socket_get_local_port() or nl_connect()).
  */
 void nl_socket_set_local_port(struct nl_sock *sk, uint32_t port)
 {
-	if (port == 0) {
-		port = generate_local_port(); 
-		/*
-		 * Release local port after generation of a new one to be
-		 * able to change local port using nl_socket_set_local_port(, 0)
-		 */
-		if (!(sk->s_flags & NL_OWN_PORT))
-			release_local_port(sk->s_local.nl_pid);
-		else
-			sk->s_flags &= ~NL_OWN_PORT;
-	} else  {
-		if (!(sk->s_flags & NL_OWN_PORT))
-			release_local_port(sk->s_local.nl_pid);
-		sk->s_flags |= NL_OWN_PORT;
-	}
-
+	if (!(sk->s_flags & NL_OWN_PORT))
+		release_local_port(sk->s_local.nl_pid);
+	sk->s_flags |= NL_OWN_PORT;
 	sk->s_local.nl_pid = port;
 }
 
@@ -339,7 +435,11 @@ int nl_socket_add_memberships(struct nl_sock *sk, int group, ...)
 		err = setsockopt(sk->s_fd, SOL_NETLINK, NETLINK_ADD_MEMBERSHIP,
 						 &group, sizeof(group));
 		if (err < 0) {
+			char buf[64];
+
 			va_end(ap);
+			NL_DBG(4, "nl_socket_add_memberships(%p): setsockopt() failed with %d (%s)\n",
+				sk, errno, strerror_r(errno, buf, sizeof(buf)));
 			return -nl_syserr2nlerr(errno);
 		}
 
@@ -387,7 +487,11 @@ int nl_socket_drop_memberships(struct nl_sock *sk, int group, ...)
 		err = setsockopt(sk->s_fd, SOL_NETLINK, NETLINK_DROP_MEMBERSHIP,
 						 &group, sizeof(group));
 		if (err < 0) {
+			char buf[64];
+
 			va_end(ap);
+			NL_DBG(4, "nl_socket_drop_memberships(%p): setsockopt() failed with %d (%s)\n",
+				sk, errno, strerror_r(errno, buf, sizeof(buf)));
 			return -nl_syserr2nlerr(errno);
 		}
 
@@ -470,6 +574,120 @@ int nl_socket_get_fd(const struct nl_sock *sk)
 	return sk->s_fd;
 }
 
+/**
+ * Set the socket file descriptor externally which initializes the
+ * socket similar to nl_connect().
+ *
+ * @arg sk         Netlink socket (required)
+ * @arg protocol   The socket protocol (optional). Linux 2.6.32 supports
+ *                 the socket option SO_PROTOCOL. In this case, you can set
+ *                 protocol to a negative value and let it autodetect.
+ *                 If you set it to a non-negative value, the detected protocol
+ *                 must match the one provided.
+ *                 To support older kernels, you must specify the protocol.
+ * @arg fd         Socket file descriptor to use (required)
+ *
+ * Set the socket file descriptor. @fd must be valid and bind'ed.
+ *
+ * This is an alternative to nl_connect(). nl_connect() creates, binds and
+ * sets the socket. With this function you can set the socket to an externally
+ * created file descriptor.
+ *
+ * @see nl_connect()
+ *
+ * @return 0 on success or a negative error code. On error, @fd is not closed but
+ * possibly unusable.
+ *
+ * @retval -NLE_BAD_SOCK Netlink socket is already connected
+ * @retval -NLE_INVAL Socket is of unexpected type
+ */
+int nl_socket_set_fd(struct nl_sock *sk, int protocol, int fd)
+{
+	int err = 0;
+	socklen_t addrlen;
+	char buf[64];
+	struct sockaddr_nl local = { 0 };
+	int so_type = -1, so_protocol = -1;
+
+	if (sk->s_fd != -1)
+		return -NLE_BAD_SOCK;
+	if (fd < 0)
+		return -NLE_INVAL;
+
+	addrlen = sizeof(local);
+	err = getsockname(fd, (struct sockaddr *) &local,
+	                  &addrlen);
+	if (err < 0) {
+		NL_DBG(4, "nl_socket_set_fd(%p,%d): getsockname() failed with %d (%s)\n",
+		       sk, fd, errno, strerror_r(errno, buf, sizeof(buf)));
+		return -nl_syserr2nlerr(errno);
+	}
+	if (addrlen != sizeof(local))
+		return -NLE_INVAL;
+	if (local.nl_family != AF_NETLINK) {
+		NL_DBG(4, "nl_socket_set_fd(%p,%d): getsockname() returned family %d instead of %d (AF_NETLINK)\n",
+		       sk, fd, local.nl_family, AF_NETLINK);
+		return -NLE_INVAL;
+	}
+
+	addrlen = sizeof(so_type);
+	err = getsockopt(fd, SOL_SOCKET, SO_TYPE, &so_type, &addrlen);
+	if (err < 0) {
+		NL_DBG(4, "nl_socket_set_fd(%p,%d): getsockopt() for SO_TYPE failed with %d (%s)\n",
+		       sk, fd, errno, strerror_r(errno, buf, sizeof(buf)));
+		return -nl_syserr2nlerr(errno);
+	}
+	if (addrlen != sizeof(so_type))
+		return -NLE_INVAL;
+	if (so_type != SOCK_RAW) {
+		NL_DBG(4, "nl_socket_set_fd(%p,%d): getsockopt() returned SO_TYPE %d instead of %d (SOCK_RAW)\n",
+		       sk, fd, so_type, SOCK_RAW);
+		return -NLE_INVAL;
+	}
+
+#if SO_PROTOCOL
+	addrlen = sizeof(so_protocol);
+	err = getsockopt(fd, SOL_SOCKET, SO_PROTOCOL, &so_protocol, &addrlen);
+	if (err < 0) {
+		if (errno == ENOPROTOOPT)
+			goto no_so_protocol;
+		NL_DBG(4, "nl_socket_set_fd(%p,%d): getsockopt() for SO_PROTOCOL failed with %d (%s)\n",
+		       sk, fd, errno, strerror_r(errno, buf, sizeof(buf)));
+		return -nl_syserr2nlerr(errno);
+	}
+	if (addrlen != sizeof(so_protocol))
+		return -NLE_INVAL;
+	if (protocol >= 0 && protocol != so_protocol) {
+		NL_DBG(4, "nl_socket_set_fd(%p,%d): getsockopt() for SO_PROTOCOL returned %d instead of %d\n",
+		       sk, fd, so_protocol, protocol);
+		return -NLE_INVAL;
+	}
+
+	if (0)
+#endif
+	{
+no_so_protocol:
+		if (protocol < 0) {
+			NL_DBG(4, "nl_socket_set_fd(%p,%d): unknown protocol and unable to detect it via SO_PROTOCOL socket option\n",
+			       sk, fd);
+			return -NLE_INVAL;
+		}
+		so_protocol = protocol;
+	}
+
+	if (sk->s_local.nl_pid != local.nl_pid) {
+		/* the port id differs. The socket is using a port id not managed by
+		 * libnl, hence reset the local port id to release a possibly generated
+		 * port. */
+		nl_socket_set_local_port (sk, local.nl_pid);
+	}
+	sk->s_local = local;
+	sk->s_fd = fd;
+	sk->s_proto = so_protocol;
+
+	return 0;
+}
+
 /**
  * Set file descriptor of socket to non-blocking state
  * @arg sk		Netlink socket.
@@ -481,8 +699,13 @@ int nl_socket_set_nonblocking(const struct nl_sock *sk)
 	if (sk->s_fd == -1)
 		return -NLE_BAD_SOCK;
 
-	if (fcntl(sk->s_fd, F_SETFL, O_NONBLOCK) < 0)
+	if (fcntl(sk->s_fd, F_SETFL, O_NONBLOCK) < 0) {
+		char buf[64];
+
+		NL_DBG(4, "nl_socket_set_nonblocking(%p): fcntl() failed with %d (%s)\n",
+			sk, errno, strerror_r(errno, buf, sizeof(buf)));
 		return -nl_syserr2nlerr(errno);
+	}
 
 	return 0;
 }
@@ -519,6 +742,9 @@ struct nl_cb *nl_socket_get_cb(const struct nl_sock *sk)
 
 void nl_socket_set_cb(struct nl_sock *sk, struct nl_cb *cb)
 {
+        if (cb == NULL)
+                BUG();
+
 	nl_cb_put(sk->s_cb);
 	sk->s_cb = nl_cb_get(cb);
 }
@@ -578,6 +804,7 @@ int nl_socket_modify_err_cb(struct nl_sock *sk, enum nl_cb_kind kind,
 int nl_socket_set_buffer_size(struct nl_sock *sk, int rxbuf, int txbuf)
 {
 	int err;
+	char buf[64];
 
 	if (rxbuf <= 0)
 		rxbuf = 32768;
@@ -590,15 +817,19 @@ int nl_socket_set_buffer_size(struct nl_sock *sk, int rxbuf, int txbuf)
 	
 	err = setsockopt(sk->s_fd, SOL_SOCKET, SO_SNDBUF,
 			 &txbuf, sizeof(txbuf));
-	if (err < 0)
+	if (err < 0) {
+		NL_DBG(4, "nl_socket_set_buffer_size(%p): setsockopt() failed with %d (%s)\n",
+			sk, errno, strerror_r(errno, buf, sizeof(buf)));
 		return -nl_syserr2nlerr(errno);
+	}
 
 	err = setsockopt(sk->s_fd, SOL_SOCKET, SO_RCVBUF,
 			 &rxbuf, sizeof(rxbuf));
-	if (err < 0)
+	if (err < 0) {
+		NL_DBG(4, "nl_socket_set_buffer_size(%p): setsockopt() failed with %d (%s)\n",
+			sk, errno, strerror_r(errno, buf, sizeof(buf)));
 		return -nl_syserr2nlerr(errno);
-
-	sk->s_flags |= NL_SOCK_BUFSIZE_SET;
+	}
 
 	return 0;
 }
@@ -649,8 +880,13 @@ int nl_socket_set_passcred(struct nl_sock *sk, int state)
 
 	err = setsockopt(sk->s_fd, SOL_SOCKET, SO_PASSCRED,
 			 &state, sizeof(state));
-	if (err < 0)
+	if (err < 0) {
+		char buf[64];
+
+		NL_DBG(4, "nl_socket_set_passcred(%p): setsockopt() failed with %d (%s)\n",
+			sk, errno, strerror_r(errno, buf, sizeof(buf)));
 		return -nl_syserr2nlerr(errno);
+	}
 
 	if (state)
 		sk->s_flags |= NL_SOCK_PASSCRED;
@@ -676,8 +912,13 @@ int nl_socket_recv_pktinfo(struct nl_sock *sk, int state)
 
 	err = setsockopt(sk->s_fd, SOL_NETLINK, NETLINK_PKTINFO,
 			 &state, sizeof(state));
-	if (err < 0)
+	if (err < 0) {
+		char buf[64];
+
+		NL_DBG(4, "nl_socket_recv_pktinfo(%p): setsockopt() failed with %d (%s)\n",
+			sk, errno, strerror_r(errno, buf, sizeof(buf)));
 		return -nl_syserr2nlerr(errno);
+	}
 
 	return 0;
 }
